#include "debugapi.h"
#include <stdarg.h>
#include <execinfo.h>
static void openlogfile();
static FILE *logfile = NULL;
#ifdef BUILD_RELEASE
bool bverbose = false;
#endif
void critical(const char *message, ...)
{
    char buf[DBG_MAX_CRITICAL_MESSAGE];
    buf[DBG_MAX_CRITICAL_MESSAGE - 1] = '\0';
    va_list va;
    va_start(va, message);
    vsnprintf(buf, DBG_MAX_CRITICAL_MESSAGE - 1, message, va);
    va_end(va);
    printf("Critical failure occured : %s\n", buf);
    if(logfile != NULL)
    {
        void *data[DBG_MAX_FRAMES];
        int size = backtrace(data, DBG_MAX_FRAMES);
        backtrace_symbols_fd(data, size, fileno(logfile));
        backtrace_symbols_fd(data, size, fileno(stdout));
    }
    exit(0);
}

static OCRAPI void openlogfile() //Fonction doesn't need to be exported
{
    logfile = fopen(LOG_FILE, "w+");
    if(logfile == NULL)
    {
        printf("Warning : Failed to open logfile : \"%s\"", LOG_FILE);
    }
}

int ocrprintf(const char *format, ...)
{
    int ret;
    va_list va;
    va_start(va, format);
    ret = vprintf(format, va);
    if(logfile == NULL)
    {
        openlogfile();
        if(logfile == NULL)
            return ret;
    }
    va_end(va);
    va_start(va, format);
    vfprintf(logfile, format, va);
    va_end(va);
    return ret;
}

int verbose(const char *format, ...)
{
#ifdef BUILD_RELEASE
    if(!bverbose)
        return 0;
#endif
    int ret;
    va_list va;
    va_start(va, format);
    ret = vprintf(format, va);
    if(logfile == NULL)
    {
        openlogfile();
        if(logfile == NULL)
            return ret;
    }
    va_end(va);
    va_start(va, format);
    vfprintf(logfile, format, va);
    va_end(va);
    return ret;
}


OCRAPI int ocrputs(const char *data)
{
    return ocrprintf(data);
}

OCRAPI int toint(float f)
{
    if(f - (int)f > 0.5)
        return (int)(f) + 1;
    else
        return (int)f;
}

OCRAPI timeval init_time()
{
    timeval t;
    gettimeofday(&t, NULL);
    return t;
}

OCRAPI unsigned long timediff(timeval t)
{
    timeval t2 = init_time();
    return (double)(t2.tv_usec - t.tv_usec) / 1000 + (t2.tv_sec - t.tv_sec) * 1000;
}

OCRAPI void setverbose(bool b)
{
#ifdef BUILD_RELEASE
    bverbose = b;
#endif
    b = b; //so ugly, but because of Werror, b has to be used
}

OCRAPI float MAXF(float f1, float f2)
{
    if(f1 > f2)
        return f1;
    return f2;
}

OCRAPI float MINF(float f1, float f2)
{
    if(f1 < f2)
        return f1;
    return f2;
}

OCRAPI int MAX(int f1, int f2)
{
    if(f1 > f2)
        return f1;
    return f2;
}

OCRAPI int MIN(int f1, int f2)
{
    if(f1 < f2)
        return f1;
    return f2;
}

OCRAPI float ABS(float f)
{
    if(f >= 0)
        return f;
    return -f;
}
