/// File : filters.h
///Purpose : Exports filtering fonctions
///like grayscale, binarization, blur, etc

#ifndef FILTERS_H
#define FILTERS_H

#include "ocrdefs.h"
#include "imgmanip.h"

#define THRESHOLD 0
#define OTSU1 1
#define OTSU2 2

Bin_Image *binarize(_in_ Image *img, int bin_type, int threshold);
/* Use the appropriate function to binarize an image, based on the bin_type
 * which must follow the value THRESHOLD, OTSU1 and OTSU2
 * If THRESHOLD is given, then the parameter threshold will be given to
 * apply_threshold
 */

Bin_Image *apply_threshold(_in_ Gray_Image*, int threshold);
// Returns a binarized version of the image

Bin_Image *apply_otsu(_in_ Gray_Image *img);
// Version of http://zerocool.is-a-geek.net/java-image/binarization/

Bin_Image *apply_otsu2(_in_ Gray_Image *img);
// Version of http://dandiggins.co.uk/arlib-9.html

Gray_Image *apply_grayscale(_in_ Image*);
// Returns a grayscale version of the image

Image *apply_gaussblur(_in_ Image*, USHORT length);
/* Applies a gaussian blur to the image, using length to define
convolution matrix's size this fonction is safe if the SDL_Surface
is valid */

#endif
