#include "neurone.h"

#define WEIGHT_COEFF (double)0.5
#define LEARN_COEFF (double)0.075;

static struct network *new_network(int nb_layer, int *size_layer)
{
    struct network *n = malloc(sizeof(struct network));
    int i;

    n->nb_layer = nb_layer;
    n->size_layer = malloc(sizeof(int) * n->nb_layer);
    n->i_layer = malloc(sizeof(int) * n->nb_layer);

    n->nb_neurone = 0;
    for (i = 0; i < nb_layer; i++)
    {
        n->size_layer[i] = size_layer[i];
        n->i_layer[i] = n->nb_neurone;
        n->nb_neurone += size_layer[i];
    }

    n->neurone = malloc(sizeof(double) * n->nb_neurone);
    n->err_neurone = malloc(sizeof(double) * n->nb_neurone);

    n->size_layer_in = size_layer[0];
    n->size_layer_out = size_layer[nb_layer-1];
    n->neurone_in = n->neurone;
    n->neurone_out = &n->neurone[n->i_layer[nb_layer-1]];
    n->err_neurone_out = &n->err_neurone[n->i_layer[nb_layer-1]];

    n->i_synapse = malloc(sizeof(int) * (n->nb_layer-1));
    n->nb_synapse = 0;
    for (i = 0; i < nb_layer - 1; i++)
    {
        n->i_synapse[i] = n->nb_synapse;
        n->nb_synapse += (n->size_layer[i]+1) * n->size_layer[i+1];
    }

    n->synapse = malloc(sizeof(double) * n->nb_synapse);

    for (i = 0; i < n->nb_synapse; i++)
    {
        n->synapse[i] = WEIGHT_COEFF * (double)rand() / RAND_MAX
            - WEIGHT_COEFF/2;

    }
    return n;
}

static void free_network(struct network *n)
{
    free(n->size_layer);
    free(n->i_layer);
    free(n->neurone);
    free(n->err_neurone);
    free(n->i_synapse);
    free(n->synapse);
    free(n);
}


static void network_treatment(struct network *n, double *inputs)
{
    int i,j, k, i_synapse=0;
    double s;

    for (i = 0; i < n->size_layer_in; i++)
            n->neurone_in[i] = inputs[i];

    for (i = 1; i < n->nb_layer; i++)
    {
        for (j = n->i_layer[i]; j < n->i_layer[i] + n->size_layer[i]; j++)
        {
            s = 0.0;
            for (k = n->i_layer[i-1];
                k < n->i_layer[i-1] + n->size_layer[i-1]; k++)
            {
                s += n->neurone[k] * n->synapse[i_synapse];
                i_synapse++;
            }
            s += n->synapse[i_synapse];
            i_synapse++;
            n->neurone[j] = s;
            if (i != n->nb_layer - 1)
                n->neurone[j] = tanh(n->neurone[j]);
        }
    }
}


static inline void get_outputs (struct network *n, double *outputs) {
    int i;
    for (i = 0; i < n->size_layer_out; i++)
    {
        outputs[i] = n->neurone_out[i];
    }
}

static void back_prop(struct network *n, double *outputs)
{
    int i,j, jj, k, i_synapse2;
    int i_synapse = n->i_synapse[n->nb_layer-2];
    double new_weight;

    for (i = 0; i < n->size_layer_out; i++)
    {
        n->err_neurone_out[i] = n->neurone_out[i] - outputs[i];
        for (j = n->i_layer[n->nb_layer-2]; j < n->i_layer[n->nb_layer-2] +
            n->size_layer[n->nb_layer-2]; j++)
        {
            new_weight = n->err_neurone_out[i] * n->neurone[j] * LEARN_COEFF;
            n->synapse[i_synapse] -= new_weight;

            if (n->synapse[i_synapse] > 5) n->synapse[i_synapse] = 5;
            if (n->synapse[i_synapse] < -5) n->synapse[i_synapse] = -5;

            i_synapse++;
        }
        new_weight = n->err_neurone_out[i] * LEARN_COEFF;
        n->synapse[i_synapse] -= new_weight;

        if (n->synapse[i_synapse] > 5)
            n->synapse[i_synapse] = 5;
        if (n->synapse[i_synapse] < -5)
            n->synapse[i_synapse] = -5;

        i_synapse++;
    }


    for (i = n->nb_layer - 2; i > 0; i--)
    {
        jj= 0;
        i_synapse = n->i_synapse[i-1];

        for (j = n->i_layer[i]; j < n->i_layer[i] +
            n->size_layer[i]; j++,jj++)
        {
            i_synapse2 = n->i_synapse[i] + jj;
            n->err_neurone[j] = 0;

            for (k = n->i_layer[i+1]; k < n->i_layer[i+1] +
                n->size_layer[i+1]; k++)
            {
                n->err_neurone[j] += n->synapse[i_synapse2] *
                    n->err_neurone[k];
                i_synapse2+=n->size_layer[i]+1;
            }

            for (k = n->i_layer[i-1]; k < n->i_layer[i-1] +
                n->size_layer[i-1]; k++)
            {
                new_weight = 1.0 - n->neurone[j] * n->neurone[j];
                new_weight *= n->err_neurone[j] * LEARN_COEFF;
                new_weight *= n->neurone[k];

                n->synapse[i_synapse] -= new_weight;
                i_synapse++;
            }

            new_weight = 1.0 - n->neurone[j] * n->neurone[j];
           new_weight *= n->err_neurone[j] * LEARN_COEFF;
            n->synapse[i_synapse] -= new_weight;
            i_synapse++;
        }
    }
}

static void xor_learning(int nb)
{
    double values[4][3] =
    {{0, 0,   0},
     {0, 1,   1},
     {1, 0,   1},
     {1, 1,   0}
    };

    double outputs;

    int size_layer[] = {2, 4, 1};
    struct network * n = new_network (3, size_layer);

    int i,k;
    for (i=0; i <= nb; i++)
    {
        k = 4.0*rand()/(RAND_MAX+1.0);
        network_treatment(n,&values[k][0]);
        back_prop(n,&values[k][2]);

        printf("=====%d/%d====\n", i, nb);

        network_treatment(n,&values[0][0]); get_outputs(n,&outputs);
        printf("0 xor 0 = \t %f\n", outputs);

        network_treatment(n,&values[1][0]); get_outputs(n,&outputs);
        printf("0 xor 1 = \t %f\n", outputs);

        network_treatment(n,&values[2][0]); get_outputs(n,&outputs);
        printf("1 xor 0 = \t %f\n", outputs);

        network_treatment(n,&values[3][0]); get_outputs(n,&outputs);
        printf("1 xor 1 = \t %f\n", outputs);

        printf("\n");
    }

    free_network(n);
}

void xor(int nb)
{
    xor_learning(nb);
}

void xor_default()
{
    xor_learning(3500);
}

