/// File : ocrdefs.h
///Purpose :  Defines global constants and
///structures for the whole project
///It should be included by any files

#ifndef OCRDEFS_H
#define OCRDEFS_H

#define _XOPEN_SOURCE 700
#define _USE_MATH_DEFINES

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>

#define ZeroMemory(addr,size) memset(addr, 0, size)
#define _in_
#define _out_
#define _inout_
#define _opt_
#define _inopt_
#define _outopt_
#define _inoutopt_
#define FILE_BUFFER 4096
#define MASK_RED 0xff0000
#define MASK_GREEN 0x00ff00
#define MASK_BLUE 0x0000ff
#define GAUSS_BLUR_RADIUS 4
#define LOG_FILE "logs.txt"
#define DBG_MAX_FRAMES 50
#define DBG_MAX_CRITICAL_MESSAGE 1024
#ifdef BUILD_RELEASE
#define OCRAPI inline
#else
#ifdef BUILD_DEBUG
#define OCRAPI
#include "debugapi.h"
#else
#error target build not specified by define (BUILD_DEBUG/BUILD_RELEASE)
#endif
#endif

#define apply_binarization(p) apply_threshold(p, 127)
#define DEFAULT_LETTERCOLOR (Pixel){0, 255, 0}
#define DEFAULT_LINECOLOR (Pixel){255, 0, 0}
#define DEFAULT_BLOCKCOLOR (Pixel){0, 0, 255}
#define RLSA_C 6
#define OPERATION_OCR 0
#define OPERATION_ROTATE 1
#define OPERATION_BINARIZE 2
#define OPERATION_RSLA 3

typedef unsigned short USHORT;
typedef unsigned int UINT;
typedef unsigned long DWORD;
typedef unsigned char BYTE;

int toint(float);
float MAXF(float, float);
float MINF(float, float);
int MAX(int,int);
int MIN(int,int);
float ABS(float);

typedef struct
{
    BYTE red;
    BYTE green;
    BYTE blue;
} Pixel;

typedef struct
{
    USHORT w;
    USHORT h;
    BYTE *data;
} Gray_Image;

typedef struct
{
    USHORT w;
    USHORT h;
    bool *data;
} Bin_Image;

typedef struct
{
    USHORT w;
    USHORT h;
    Pixel *data;
}Image ;

typedef struct
{
    USHORT x;
    USHORT y;
    USHORT w;
    USHORT h;
} Rect ;

typedef struct
{
    Rect pos;
    Pixel color;
    char value;
} Letter;

#endif // OCRDEFS_H
