#include "textextraction.h"
///internal fonctions

static USHORT nextline(_inout_ Queue*, _in_ Bin_Image *img, USHORT line,
                       Rect r);
static USHORT nextchar(_inout_ Queue*, _in_ Bin_Image *img, _in_ Rect *line,
                       USHORT c);

static bool is_subcolumnblank(_in_ Bin_Image *img, USHORT x, USHORT y,
                              USHORT h);
static bool is_sublineblank(_in_ Bin_Image *img, USHORT x, USHORT u, USHORT w);

static void adjust_letterrect(_in_ Bin_Image *img, _inout_ Rect*);
static void adjust_linerect(_in_ Bin_Image *img, _inout_ Rect*);
static void adjust_blockrect(_inout_ Rect *r, USHORT w, USHORT h);

static Letter *finalizedetection(_in_ Image *img, _in_ Bin_Image* bimg,
                                 USHORT woffset, USHORT hoffset, _in_ Queue *q, _out_ USHORT *count,
                                 bool colors);
static void finalizelines(USHORT woffset, USHORT hoffset, _inout_ Rect* lines,
                          USHORT lcount);

static Pixel getrectcolor(_in_ Image *img, _in_ Bin_Image *binmg, Rect r,
                          USHORT woffset, USHORT hoffset);

static Queue *extract_blocs(_in_ Bin_Image *img, USHORT rx, USHORT ry);
static bool nextblock(_inout_ Bin_Image *img, _inout_ Queue *queue);
static bool truepixelnear(_in_ Bin_Image *img, USHORT x, USHORT y, USHORT rx,
                          USHORT ry);
static void adjust_block(_in_ Bin_Image *img, Rect *r);
static void get_rlsavalues(_in_ Bin_Image *img, _out_ USHORT *w,
                           _out_ USHORT *h);
static void rlsa_computeline(_in_ Bin_Image *img, Rect rect, _out_ DWORD *tx,
                             _out_ DWORD *cx);
static void rlsa_computeblock(_in_ Bin_Image *img, Rect rect, _out_ DWORD *ty,
                              _out_ DWORD *cy);

static USHORT irlsaradius = RLSA_C;
static USHORT irlsaforcedradius = 0;

OCRAPI void extract_lines(Bin_Image *img, Rect rect, Queue * queue)
{
    USHORT line = rect.y;
    while(line < rect.y + rect.h)
    {
        line = nextline(queue, img, line, rect);
    }
}

static USHORT nextline(Queue *q, Bin_Image *img, USHORT line, Rect rect)
{
    line--;
    do
    {
        line++;
        if(line == rect.y + rect.h)
            return line;
    }
    while(islineblank(img, line));
    USHORT start = line;
    line++;
    while(line < rect.y + rect.h && !islineblank(img, line))
    {
        line++;
    }
    Rect *r = malloc(sizeof(Rect));
    r->h = line - start - 1;
    r->w = rect.w;
    r->x = rect.x;
    r->y = start;
    adjust_linerect(img, r);
    push(q, r);
    return line;
}

void extract_chars(Bin_Image *img, Queue *lines, Queue *letters, Rect *rects)
{
    USHORT c;
    Rect *r;
    USHORT i = 0;
    while(!isqueueempty(lines))
    {
        r = pop(lines);
        c = r->x;
        while(c < r->w + r->x)
        {
            c = nextchar(letters, img, r, c);
        }
        if(rects != NULL)
        {
            rects[i] = *r;
            i++;
        }
        free(r);
    }
    freequeue(lines);
}

static USHORT nextchar(Queue *q, Bin_Image *img, Rect *line, USHORT c)
{
    while(c < line->x + line->w && is_subcolumnblank(img, c, line->y, line->h))
    {
        c++;
        if(c == line->w)
        {
            return c;
        }
    }
    USHORT left = c;
    c++;
    while(c < line->w + line->x + 1
            && !is_subcolumnblank(img, c - 1, line->y, line->h))
    {
        c++;
    }
    Rect *r = malloc(sizeof(Rect));
    r->x = left;
    r->w = c - left - 1;
    r->y = line->y;
    r->h = line->h;
    adjust_letterrect(img, r);
    push(q, r);
    return c;
}

static OCRAPI bool is_subcolumnblank(Bin_Image *img, USHORT x, USHORT y,
                                     USHORT h)
{
    while(h > 0 && !getpixelbin(img, x, y))
    {
        y++;
        h--;
    }
    return h == 0;
}

static OCRAPI bool is_sublineblank(Bin_Image *img, USHORT x, USHORT y, USHORT w)
{
    while(w > 0 && !getpixelbin(img, x, y))
    {
        x++;
        w--;
    }
    return w == 0;
}

OCRAPI void adjust_letterrect(Bin_Image *img,
                              Rect *rect) //No need to check columns
{
    while(is_sublineblank(img, rect->x, rect->y, rect->w ) && rect->h > 0)
    {
        rect->y++;
        rect->h--;
    }

    while(is_sublineblank(img, rect->x, MIN(rect->y + rect->h, img->h - 1),
                          rect->w ) && rect->h > 0)
    {
        rect->h--;
    }
}

OCRAPI void adjust_linerect(Bin_Image *img, Rect *r)
{
    while(r->w > 0 && is_subcolumnblank(img, r->x, r->y, r->h))
    {
        r->x++;
        r->w--;
    }
    while(r->w > 0
            && is_subcolumnblank(img, MIN(r->x + r->w, img->w - 1), r->y, r->h))
    {
        r->w--;
    }
}

static Pixel getrectcolor(Image *img, Bin_Image *bimg, Rect rect,
                          USHORT woffset, USHORT hoffset)
{
    USHORT x;
    USHORT y;
    DWORD r = 0;
    DWORD g = 0;
    DWORD b = 0;
    DWORD count = 0;
    Pixel p;
    for(x = rect.x; x < rect.x + rect.w; x++)
    {
        for(y = rect.y; y < rect.y + rect.h; y++)
        {
            if(getpixelbin(bimg, x - woffset, y - hoffset))
            {
                p = getpixel(img, x, y);
                r += p.red;
                g += p.green;
                b += p.blue;
                count++;
            }
        }
    }
    p.red = toint(r / (float)count);
    p.green = toint(g / (float)count);
    p.blue = toint(b / (float)count);
    return p;
}

static Letter *finalizedetection(Image *img, Bin_Image *bimg, USHORT woffset,
                                 USHORT hoffset, Queue *q, USHORT *count, bool colors)
{
    *count = 0;
    Letter *ret = malloc(sizeof(Letter) * q->count);
    Rect *r;
    while(!isqueueempty(q))
    {
        r = pop(q);
        ret[*count].pos.x = r->x + woffset;
        ret[*count].pos.y = r->y + hoffset;
        ret[*count].pos.w = r->w;
        ret[*count].pos.h = r->h;
        if(colors)
            ret[*count].color = getrectcolor(img, bimg, ret[*count].pos, woffset, hoffset);
        else
            ret[*count].color = (Pixel)
        {
            0, 0, 0
        };
        ret[*count].value = '\0';
        free(r);
        (*count)++;
    }
    freequeue(q);
    return ret;
}

static OCRAPI void adjust_blockrect(Rect *r, USHORT w, USHORT h)
{
    r->x += w;
    r->y += h;
}

static Queue *extract_blocs( Bin_Image *img, USHORT rx, USHORT ry)
{
    Bin_Image *buf = create_bin_image(img->w, img->h);
    memcpy(buf->data,img->data,sizeof(bool)*img->w*img->h);
    USHORT x;
    USHORT y;
    for(x = 0; x < img->w; x++)
    {
        for(y = 0; y < img->h; y++)
        {
            if(!getpixelbin(img, x, y) && truepixelnear(img, x, y, rx, ry))
                setpixelbin(buf, x, y, true);
        }
    }
    Queue *q = createqueue();

    while(true)
    {
        if(!nextblock(buf, q))
            break;
    }
    free_bimage(buf);
    return q;
}

static bool nextblock(Bin_Image *img, Queue *q)
{
    USHORT x;
    USHORT y;
    for(x = 0; x < img->w; x++)
    {
        for(y = 0; y < img->h; y++)
        {
            if(getpixelbin(img, x, y))
            {
                Rect *r = malloc(sizeof(Rect));
                r->x = x;
                r->y = y;
                r->w = 1;
                r->h = 1;
                while(x + r->w + 1 < img->w && getpixelbin(img, x + r->w + 1, y))
                {
                    r->w++;
                }
                while(y + r->h + 1 < img->h && getpixelbin(img, x, y + r->h + 1))
                {
                    r->h++;
                }
                adjust_block(img, r);
                push(q, r);
                for(x = r->x; x < r->x + r->w; x++)
                {
                    for(y = r->y; y < r->y + r->h; y++)
                    {
                        setpixelbin(img, x, y, false);
                    }
                }
                return true;
            }
        }
    }
    return false;
}

static void adjust_block(Bin_Image *img, Rect *r)
{
    bool b;
    USHORT x;
    USHORT y;
    do
    {
        b = false;
        for(x = r->x; x < r->w + r->x; x++)
        {
            while(r->y && getpixelbin(img, x, r->y - 1))
            {
                r->y--;
                r->h++;
                b = true;
            }
            while(r->y + r->h < img->h && getpixelbin(img, x, r->y + r->h))
            {
                r->h++;
                b = true;
            }
        }
        for(y = r->y; y < r->h + r->y; y++)
        {
            while(r->x && getpixelbin(img, r->x - 1, r->y))
            {
                r->x--;
                r->w++;
                b = true;
            }

            while(r->x + r->w < img->w  && getpixelbin(img, r->x + r->w, y))
            {
                r->w++;
                b = true;
            }
        }
    }
    while(b);
}
static OCRAPI bool truepixelnear(Bin_Image *img, USHORT x, USHORT y, USHORT rx,
                                 USHORT ry)
{
    USHORT cx;
    USHORT cy;
    for(cx = MAX(x - rx, 0); cx < x + rx && cx < img->w; cx++)
    {
        for(cy = MAX(y - rx, 0); cy < y + ry && cy < img->h; cy++)
        {
            if(getpixelbin(img, cx, cy))
                return true;
        }
    }
    return false;
}
static OCRAPI void finalizelines(USHORT w, USHORT h, Rect *lines, USHORT lc)
{
    USHORT i;
    for(i = 0; i < lc; i++)
    {
        lines[i].x += w;
        lines[i].y += h;
    }
}
OCRAPI static void rlsa_computeline(Bin_Image *img, Rect r, DWORD *total,
                                    DWORD *count)
{
    USHORT x = r.x;
    USHORT start;
    while(x < r.x + r.h)
    {
        while(x < r.w + r.x && !is_subcolumnblank(img, x, r.y, r.h))
        {
            x++;
        }
        if(x == r.w + r.x)
            break;
        start = x;
        while(x < r.w + r.x)
        {
            if(!is_subcolumnblank(img, x, r.y, r.h))
            {
                *total += x - start;
                (*count)++;
                break;
            }
            x++;
        }
    }
}

OCRAPI static void rlsa_computeblock(Bin_Image *img, Rect r, DWORD *total,
                                     DWORD *count)
{
    USHORT y = r.y;
    USHORT start;
    while(y < r.y + r.h)
    {
        while(y < r.h + r.y && !is_sublineblank(img, r.x, y, r.w))
        {
            y++;
        }
        if(y == r.y + r.h)
            break;
        start = y;
        while(y < r.h + r.y)
        {
            if(!is_sublineblank(img, r.x, y, r.w))
            {
                *total += y - start;
                (*count)++;
                break;
            }
            y++;
        }
    }
}
static void get_rlsavalues(Bin_Image *img, USHORT *w, USHORT *h)
{
    DWORD totalx = 0;
    DWORD countx = 0;
    DWORD totaly = 0;
    DWORD county = 0;
    Queue *blocks = extract_blocs(img, irlsaradius, irlsaradius);
    Queue *lines = createqueue();
    Rect *rect;
    while(!isqueueempty(blocks))
    {
        rect = pop(blocks);
        rlsa_computeblock(img, *rect, &totaly, &county);
        extract_lines(img, *rect, lines);
        free(rect);
    }
    freequeue(blocks);
    while(!isqueueempty(lines))
    {
        rect = pop(lines);
        rlsa_computeline(img, *rect, &totalx, &countx);
        free(rect);
    }
    freequeue(lines);
    *w = MAX(toint(totalx / (float)countx) * 4, 5);
    *h = toint(totaly / (float)county) + 3;
}
OCRAPI void setrlsa_radius(USHORT r)
{
    irlsaradius = r;
}

OCRAPI void forcerlsa_radius(USHORT r)
{
    irlsaforcedradius = r;
}
void detect_anddraw(Image *img, bool colors, int btype, int threshold)
{
    timeval t = init_time();
    USHORT linecount;
    USHORT lettercount;
    Rect *rect;
    USHORT rslax;
    USHORT rslay;
    USHORT w = img->w;
    USHORT h = img->h;
    Bin_Image *bin = binarize(img, btype, threshold);
    verbose("Binarized (%lu ms)\n", timediff(t));

    remove_blank(bin, &w, &h);
    verbose("Removed blank (%lu ms)\n", timediff(t));
    if(irlsaforcedradius)
    {
        rslax = irlsaforcedradius;
        rslay = irlsaforcedradius;
        printf("Using forced rlsa value : %u\n", rslax);
    }
    else
    {
        get_rlsavalues(bin, &rslax, &rslay);
        verbose("Computed rsla values (%u, %u) (%lu ms)\n", rslax, rslay, timediff(t));
    }

    Queue *blocks = extract_blocs(bin, rslax, rslay);
    verbose("Extracted text blocs (%lu ms)\n", timediff(t));
    Queue *lines = createqueue();
    while(!isqueueempty(blocks))
    {
        rect = pop(blocks);
        extract_lines(bin, *rect, lines);
        adjust_blockrect(rect, w, h);
        drawfromrect(img, *rect, DEFAULT_BLOCKCOLOR);
        free(rect);
    }
    verbose("Extracted lines (%lu ms)\n", timediff(t));

    freequeue(blocks);
    Queue *chars = createqueue();
    linecount = lines->count;
    Rect *linerect = malloc(sizeof(Rect) * linecount);
    extract_chars(bin, lines, chars, linerect);
    verbose("Extracted characters (%lu ms)\n", timediff(t));

    Letter *letters = finalizedetection(img, bin, w, h, chars, &lettercount,
                                        colors);
    finalizelines(w, h, linerect, linecount);
    verbose("Finalized detection (%lu ms)\n", timediff(t));

    for(w = 0; w < lettercount; w++)
    {
        if(colors)
            drawfromrect(img, letters[w].pos, letters[w].color);
        else
            drawfromrect(img, letters[w].pos, DEFAULT_LETTERCOLOR);
    }
    for(w = 0; w < linecount; w++)
    {
        drawfromrect(img, linerect[w], DEFAULT_LINECOLOR);
    }
    verbose("Drawn rectangles (%lu ms)\n", timediff(t));

    free_bimage(bin);
    free(letters);
    free(linerect);
    verbose("Cleaned up (%lu ms)\n", timediff(t));
}
